/* Frequency counter module for PC
 * F is devided by external counter e.q. 74HC4040 to a frequency below 200 Hz
 * The output of counter is connected to LPT port pin 10 (!ACK)
 * Programs reads 8254 counter on every LPT interrupt and calculates
 * frequency from period between 2 interrupts. These values are store
 * in the 'periods' array.
 * Number of measurements is set by specifying a measurement period
 * (gate time). The timeout is twice this value.
 */

#include <stdio.h>
#include <dos.h>
#include <conio.h>
#include "vars.h"
#include "freqcnt.h"

#define	PSIZE		1024		/* Length of periods array	*/
#define	CLK8254		1193180.0	/* Counter clock in Hz		*/
#define	MAXERR		0.01		/* 1 % maximum error		*/

static void open_parport(void);
static void close_parport(void);
static void interrupt (*get_irq_vec(int));
static void set_irq_vec(int, void interrupt());
static void enable_irq(int);
static void disable_irq(int);
static void interrupt lpt_intr(void);
static void interrupt (*oldlpt_intr)(void);

static unsigned long periods[PSIZE];	/* measurements buffer	*/
static int pindex;			/* index in above or -1	*/
static unsigned long laststamp;		/* last valid time stamp */

/* Start measurement.
 * Enables LPT interrupt
 * Clears index into periods array
 */

static unsigned int tick0;		/* Starting tick		*/
static unsigned int gateticks;		/* # ticks to go		*/
static unsigned int to_ticks;		/* # ticks to go before timeout	*/
static int busy;			/* Set if counting in progress	*/

void
fcount_start(int tgate)
{
	/* Initialize. Skip first interrupt, it might be
	 * a spurious one
	 */
	pindex = -1;

	/* Calculate approx. number of gate ticks
	 * rounded up
	 * Ticks approx. 55 msec
	 */
	switch (tgate) {
	case GATE10MS:	gateticks = 1; break;
	case GATE100MS:	gateticks = 2; break;
	case GATE200MS:	gateticks = 4; break;
	case GATE500MS:	gateticks = 10; break;
	case GATE1S:	gateticks = 19; break;
	default:	gateticks = 19; break;	/* 1 sec */
	}

	/* Read ticks, used for timeout in fcount wait */
	tick0 = * (unsigned int far *)MK_FP(0x40,0x6c);
	to_ticks = gateticks * 2; /* allows for 50 % misses	*/

	/* Configure interrupt routine */
	open_parport();
}

/* Wait for measurements to finish
 * On timeout returns -1 else return 0
 * Lowest frequency (after division) is 1 Hz
 * Timeout is twice gateperiod
 */
unsigned int
fcount_wait(void)
{
	int tick1;		/* Current tick */

	for(;;) {
		if (busy == 0) {
			/* All done */
			break;
		}
		tick1 = * (unsigned int far *)MK_FP(0x40,0x6c);
		if (tick1 > (tick0 + to_ticks)) {
			close_parport();
			return(-1);
		}
	}
	return(0);
}

/* Measurents are store in periods array.
 * fcount_read converts n entries into a frequency
 */
double
fcount_read(void)
{
	unsigned long elapsed;
	double f;
	int i;
	unsigned long t_av;		/* Average period	*/
	long t_delta;

	/* Values available ? */
	if (pindex <= 0) {
		return -1.0;
	}

	/* Accumulate entries */
	elapsed = 0L;
	for(i = 0 ; i < pindex ; i++) {
		elapsed += periods[i];
	}

	/* Experimental
	 * Check for unexpected values
	 */
	t_av = elapsed / pindex;
	for(i = 0 ; i < pindex ; i++) {
		t_delta = periods[i] - t_av;
		if (t_delta < 0)
			t_delta = - t_delta;
		/* More dan MAXERR error ? */
		if ((t_delta / periods[i]) > MAXERR) {
			printf("%d: t_av=%ld period=%ld, ",
				i,t_av,periods[i]);
		}
	}

	/* Now we have n periods in units of
	 * ~838 nsec, fclk of 8254 is 1.193180 Mhz
	 * T = elapsed * (1 / fclk)
	 * f = n * (1 / T) * p = n * p * fclk / elapsed
	 * where p is prescaler value
	 */
	f = (double)pindex *
		(double)FC_PRESCALE *
		(double)CLK8254 /
		(double)elapsed;

	return(f);
}


/* Parallel port routines
 */

void
open_parport(void)
{
	/* Init administration */
	laststamp = -1L;

	/* Initialize LPT port
	 * Bit	function			value
	 *  7	reserved			0
	 *  6	reserved			0
	 *  5	enable bidir port 		0
	 *  4	enable IEQ			1
	 *  3	select printer			0
	 *  2	init printer			0
	 *  1	auto linefeed			0
	 *  0	strobe				0
	 */
	outportb(lpt_base + 2, 0x10);

	/* save old interrupt vector and set new interrupt vector */
	disable();

	/* enable interrupt on 8259 */
	oldlpt_intr = get_irq_vec(lpt_irq);
	set_irq_vec(lpt_irq,lpt_intr);
	enable_irq(lpt_irq);

	/* All set */
	busy = 1;
	enable();
}

void
close_parport(void)
{
	/* Turn of serial interrupt on 8259 */
	disable();
	disable_irq(lpt_irq);
	set_irq_vec(lpt_irq,oldlpt_intr);
	busy = 0;
	enable();
}

/* Return time since startup in milliseconds. If the system has an
 * 8254 clock chip (standard on ATs and up) then resolution is improved
 * below 55 ms (the clock tick interval) by reading back the instantaneous
 * value of the counter and combining it with the global clock tick counter.
 * Otherwise 55 ms resolution is provided.
 *
 * Reading the 8254 is a bit tricky since a tick could occur asynchronously
 * between the two reads. The tick counter is examined before and after the
 * hardware counter is read. If the tick counter changes, try again.
 * Note: the hardware counter counts down from 65536.
 */

/* 8254 count sequence
 *
 *	count		countervalue	status.15	x	desired-value
 *	0		0000		8000		8000	0
 *	1		fffe		8000		ffff	1
 *	2		fffc		8000		fffe	2
 *	3		fffa		8000		fffd	3
 *	..
 *	7fff		2		8000		8001	7fff
 *	8000		0		0		0	8000
 *	8001		fffe		0		7fff	8001
 *	8002		fffc		0		7ffe	8002
 *	..
 *	ffff		2		0		1	ffff
 *	0		0		8000		0	0
 *
 *	x = (countervalue >> 1) | status.15
 *	if (countervalue == 0)
 *		desired-value = status.15 ^ 0x8000
 *	else
 *		desired-value = -x
 */

void interrupt
lpt_intr(void)
{
	unsigned long tstamp;		/* Time stamp		*/
	unsigned int hi,t;
	unsigned short lo;
	unsigned short st;
	unsigned short i;

	/* Report interrupt serviced 8259 */
	if (lpt_irq >= 8)
		outportb(0xa0,0x20);
	outportb(0x20,0x20);

	/* Get and store time stamp */

	/* set hi to 'ticks-since-midnight */
        hi = * (unsigned int far *)MK_FP(0x40,0x6c);

	/* Now get current timer 0 value */
	outportb(0x43,0xc2);
	st = inportb(0x40) << 8;
	lo = inportb(0x40);
	lo = lo + (inportb(0x40) << 8);
	/* .. */
	st = st & 0x8000;
	if (lo == 0 )
		lo = st ^ 0x8000;
	else
		lo = -((lo>>1) | st);

	/* Carry out timing problem in 8254 ???? */
	if (lo == 0xffff || lo == 0x0000) {
		tstamp = -1L;
	} else {
		t = * (unsigned int far *)MK_FP(0x40,0x6c);
		if (hi != t )
			tstamp = -1L;
		else
			tstamp = ((unsigned long)hi << 16) | lo;
	}

	/* Only valid if tstamp > 0 */
	if (tstamp != -1L) {
		/* Previous value must be valid to compute
		 * valid delta
		 */
		if (laststamp != -1L) {
			/* pindex MUST be > 0, -1 signals first interrupt
			 */
			if (pindex >= 0)
				periods[pindex] = tstamp - laststamp;
			pindex++;
			/* Prevent overflow */
			if (pindex >= PSIZE)
				pindex--;
		}
		laststamp = tstamp;
	} else {
		laststamp = -1L;
	}

	/* All done */
	if ((hi - tick0) >= gateticks) {
		close_parport();
	}
}

/* Usefull subroutines */
void interrupt
(*get_irq_vec(int irq))
{
	if (irq < 8) {
		return getvect(8+irq);
	} else {
		return getvect(0x70+irq-8);
	}
}

void
set_irq_vec(int irq, void interrupt (*irqhandler)())
{
	if (irq < 8) {
		setvect(8+irq,irqhandler);
	} else {
		setvect(0x70 + irq - 8,irqhandler);
	}
}

void
enable_irq(int irq)
{
	unsigned char i;

	if (irq < 8) {
		i = inportb(0x21);
		i &= ~(1<<irq);
		outportb(0x21,i);
	} else {
		i = inportb(0xa1);
		i &= ~(1<<(irq - 8));
		outportb(0xa1,i);
	}
}

void
disable_irq(int irq)
{
	unsigned char i;

	if (irq < 8) {
		i = inportb(0x21);
		i |= (1<<irq);
		outportb(0x21,i);
	} else {
		i = inportb(0xa1);
		i |= (1<<(irq - 8));
		outportb(0xa1,i);
	}
}

